
from PySide2 import QtCore

qt_resource_data = b"\
\x00\x00\x18\x9c\
<\
?xml version=\x221.\
0\x22 encoding=\x22UTF\
-8\x22 standalone=\x22\
no\x22?>\x0a<!-- Creat\
ed with Inkscape\
 (http://www.ink\
scape.org/) -->\x0a\
<svg xmlns:dc=\x22h\
ttp://purl.org/d\
c/elements/1.1/\x22\
 xmlns:cc=\x22http:\
//creativecommon\
s.org/ns#\x22 xmlns\
:rdf=\x22http://www\
.w3.org/1999/02/\
22-rdf-syntax-ns\
#\x22 xmlns:svg=\x22ht\
tp://www.w3.org/\
2000/svg\x22 xmlns=\
\x22http://www.w3.o\
rg/2000/svg\x22 xml\
ns:xlink=\x22http:/\
/www.w3.org/1999\
/xlink\x22 xmlns:so\
dipodi=\x22http://s\
odipodi.sourcefo\
rge.net/DTD/sodi\
podi-0.dtd\x22 xmln\
s:inkscape=\x22http\
://www.inkscape.\
org/namespaces/i\
nkscape\x22 width=\x22\
64\x22 height=\x2264\x22 \
id=\x22svg2\x22 versio\
n=\x221.1\x22 inkscape\
:version=\x220.48.5\
 r10040\x22 sodipod\
i:docname=\x22previ\
ew-rendered.svg\x22\
>\x0a  <defs id=\x22de\
fs4\x22>\x0a    <linea\
rGradient id=\x22li\
nearGradient3755\
\x22>\x0a      <stop s\
tyle=\x22stop-color\
:#424242;stop-op\
acity:1;\x22 offset\
=\x220\x22 id=\x22stop375\
7\x22/>\x0a      <stop\
 style=\x22stop-col\
or:#9f9f9f;stop-\
opacity:1;\x22 offs\
et=\x221\x22 id=\x22stop3\
759\x22/>\x0a    </lin\
earGradient>\x0a   \
 <filter color-i\
nterpolation-fil\
ters=\x22sRGB\x22 inks\
cape:collect=\x22al\
ways\x22 id=\x22filter\
3871\x22 x=\x22-0.4536\
3569\x22 width=\x221.9\
072714\x22 y=\x22-0.45\
363569\x22 height=\x22\
1.9072714\x22>\x0a    \
  <feGaussianBlu\
r inkscape:colle\
ct=\x22always\x22 stdD\
eviation=\x229.8975\
056\x22 id=\x22feGauss\
ianBlur3873\x22/>\x0a \
   </filter>\x0a   \
 <linearGradient\
 inkscape:collec\
t=\x22always\x22 xlink\
:href=\x22#linearGr\
adient3600\x22 id=\x22\
linearGradient36\
06\x22 x1=\x2251.03728\
1\x22 y1=\x2247.692612\
\x22 x2=\x2214.872466\x22\
 y2=\x2210.397644\x22 \
gradientUnits=\x22u\
serSpaceOnUse\x22/>\
\x0a    <linearGrad\
ient id=\x22linearG\
radient3600\x22>\x0a  \
    <stop style=\
\x22stop-color:#555\
753;stop-opacity\
:1\x22 offset=\x220\x22 i\
d=\x22stop3602\x22/>\x0a \
     <stop style\
=\x22stop-color:#d3\
d7cf;stop-opacit\
y:1\x22 offset=\x221\x22 \
id=\x22stop3604\x22/>\x0a\
    </linearGrad\
ient>\x0a    <filte\
r inkscape:colle\
ct=\x22always\x22 id=\x22\
filter3805\x22 x=\x22-\
0.29648924\x22 widt\
h=\x221.5929785\x22 y=\
\x22-0.30309633\x22 he\
ight=\x221.6061927\x22\
>\x0a      <feGauss\
ianBlur inkscape\
:collect=\x22always\
\x22 stdDeviation=\x22\
2.2490136\x22 id=\x22f\
eGaussianBlur380\
7\x22/>\x0a    </filte\
r>\x0a  </defs>\x0a  <\
sodipodi:namedvi\
ew id=\x22base\x22 pag\
ecolor=\x22#ffffff\x22\
 bordercolor=\x22#6\
66666\x22 borderopa\
city=\x221.0\x22 inksc\
ape:pageopacity=\
\x220.0\x22 inkscape:p\
ageshadow=\x222\x22 in\
kscape:zoom=\x224.8\
193618\x22 inkscape\
:cx=\x22-4.0027486\x22\
 inkscape:cy=\x2228\
.204234\x22 inkscap\
e:document-units\
=\x22px\x22 inkscape:c\
urrent-layer=\x22la\
yer1\x22 showgrid=\x22\
true\x22 showborder\
=\x22true\x22 inkscape\
:window-width=\x228\
00\x22 inkscape:win\
dow-height=\x22837\x22\
 inkscape:window\
-x=\x220\x22 inkscape:\
window-y=\x2227\x22 in\
kscape:window-ma\
ximized=\x220\x22>\x0a   \
 <inkscape:grid \
type=\x22xygrid\x22 id\
=\x22grid2988\x22 emps\
pacing=\x222\x22 visib\
le=\x22true\x22 enable\
d=\x22true\x22 snapvis\
iblegridlinesonl\
y=\x22true\x22/>\x0a  </s\
odipodi:namedvie\
w>\x0a  <metadata i\
d=\x22metadata7\x22>\x0a \
   <rdf:RDF>\x0a   \
   <cc:Work rdf:\
about=\x22\x22>\x0a      \
  <dc:format>ima\
ge/svg+xml</dc:f\
ormat>\x0a        <\
dc:type rdf:reso\
urce=\x22http://pur\
l.org/dc/dcmityp\
e/StillImage\x22/>\x0a\
        <dc:titl\
e/>\x0a        <dc:\
creator>\x0a       \
   <cc:Agent>\x0a  \
          <dc:ti\
tle>$committer</\
dc:title>\x0a      \
    </cc:Agent>\x0a\
        </dc:cre\
ator>\x0a        <d\
c:title>preview-\
rendered</dc:tit\
le>\x0a        <dc:\
date>2013-11-19<\
/dc:date>\x0a      \
  <dc:relation>h\
ttp://www.freeca\
d.org/wiki/index\
.php?title=Artwo\
rk</dc:relation>\
\x0a        <dc:pub\
lisher>\x0a        \
  <cc:Agent>\x0a   \
         <dc:tit\
le>FreeCAD</dc:t\
itle>\x0a          \
</cc:Agent>\x0a    \
    </dc:publish\
er>\x0a        <dc:\
identifier>FreeC\
AD/src/Mod/Mater\
ial/Resources/ic\
ons/preview-rend\
ered.svg</dc:ide\
ntifier>\x0a       \
 <dc:rights>\x0a   \
       <cc:Agent\
>\x0a            <d\
c:title>FreeCAD \
LGPL2+</dc:title\
>\x0a          </cc\
:Agent>\x0a        \
</dc:rights>\x0a   \
     <cc:license\
>https://www.gnu\
.org/copyleft/le\
sser.html</cc:li\
cense>\x0a        <\
dc:contributor>\x0a\
          <cc:Ag\
ent>\x0a           \
 <dc:title>[agry\
son] Alexander G\
ryson</dc:title>\
\x0a          </cc:\
Agent>\x0a        <\
/dc:contributor>\
\x0a      </cc:Work\
>\x0a    </rdf:RDF>\
\x0a  </metadata>\x0a \
 <g inkscape:lab\
el=\x22Layer 1\x22 ink\
scape:groupmode=\
\x22layer\x22 id=\x22laye\
r1\x22 transform=\x22t\
ranslate(0,-988.\
36218)\x22>\x0a    <pa\
th sodipodi:type\
=\x22arc\x22 style=\x22co\
lor:#000000;fill\
:#000000;fill-op\
acity:0.65625;fi\
ll-rule:evenodd;\
stroke:none;stro\
ke-width:5;marke\
r:none;visibilit\
y:visible;displa\
y:inline;overflo\
w:visible;filter\
:url(#filter3871\
);enable-backgro\
und:accumulate\x22 \
id=\x22path2826-8\x22 \
sodipodi:cx=\x2231.\
636364\x22 sodipodi\
:cy=\x2230.363636\x22 \
sodipodi:rx=\x2226.\
181818\x22 sodipodi\
:ry=\x2226.181818\x22 \
d=\x22m 57.818182,3\
0.363636 a 26.18\
1818,26.181818 0\
 1 1 -52.363636,\
0 26.181818,26.1\
81818 0 1 1 52.3\
63636,0 z\x22 trans\
form=\x22matrix(0.4\
8061681,0,0,0.12\
01542,24.795032,\
1042.7139)\x22/>\x0a  \
  <path sodipodi\
:type=\x22arc\x22 styl\
e=\x22color:#000000\
;fill:url(#linea\
rGradient3606);f\
ill-opacity:1;fi\
ll-rule:evenodd;\
stroke:#0b1521;s\
troke-width:1.80\
564272;marker:no\
ne;visibility:vi\
sible;display:in\
line;overflow:vi\
sible;enable-bac\
kground:accumula\
te\x22 id=\x22path2826\
\x22 sodipodi:cx=\x223\
1.636364\x22 sodipo\
di:cy=\x2230.363636\
\x22 sodipodi:rx=\x222\
6.181818\x22 sodipo\
di:ry=\x2226.181818\
\x22 d=\x22m 57.818182\
,30.363636 a 26.\
181818,26.181818\
 0 1 1 -52.36363\
6,0 26.181818,26\
.181818 0 1 1 52\
.363636,0 z\x22 tra\
nsform=\x22matrix(1\
.1076389,0,0,1.1\
076389,-3.041667\
2,986.73024)\x22/>\x0a\
    <path sodipo\
di:type=\x22arc\x22 st\
yle=\x22color:#0000\
00;fill:none;str\
oke:#000000;stro\
ke-width:1.93939\
412000000000;mar\
ker:none;visibil\
ity:visible;disp\
lay:inline;overf\
low:visible;enab\
le-background:ac\
cumulate\x22 id=\x22pa\
th2826-4\x22 sodipo\
di:cx=\x2231.636364\
\x22 sodipodi:cy=\x223\
0.363636\x22 sodipo\
di:rx=\x2226.181818\
\x22 sodipodi:ry=\x222\
6.181818\x22 d=\x22m 5\
7.818182,30.3636\
36 a 26.181818,2\
6.181818 0 1 1 -\
52.363636,0 26.1\
81818,26.181818 \
0 1 1 52.363636,\
0 z\x22 transform=\x22\
matrix(1.03125,0\
,0,1.03125,-0.62\
500046,989.04968\
)\x22/>\x0a    <path s\
odipodi:type=\x22ar\
c\x22 style=\x22opacit\
y:0.844827569999\
99986;color:#000\
000;fill:#fffffe\
;fill-opacity:1;\
fill-rule:evenod\
d;stroke:none;st\
roke-width:5;mar\
ker:none;visibil\
ity:visible;disp\
lay:inline;overf\
low:visible;enab\
le-background:ac\
cumulate;filter:\
url(#filter3805)\
\x22 id=\x22path3608\x22 \
sodipodi:cx=\x2211.\
181818\x22 sodipodi\
:cy=\x2220.181818\x22 \
sodipodi:rx=\x2210.\
454545\x22 sodipodi\
:ry=\x227.2727275\x22 \
d=\x22m 21.636363,2\
0.181818 a 10.45\
4545,7.2727275 0\
 1 1 -20.9090900\
1,0 10.454545,7.\
2727275 0 1 1 20\
.90909001,0 z\x22 t\
ransform=\x22matrix\
(0.85867478,-0.7\
467709,0.7057652\
6,0.90856459,-4.\
8393203,997.4188\
)\x22/>\x0a    <path s\
odipodi:type=\x22ar\
c\x22 style=\x22color:\
#000000;fill:non\
e;stroke:#2e3436\
;stroke-width:1.\
8056427200000000\
3;marker:none;vi\
sibility:visible\
;display:inline;\
overflow:visible\
;enable-backgrou\
nd:accumulate\x22 i\
d=\x22path2826-5\x22 s\
odipodi:cx=\x2231.6\
36364\x22 sodipodi:\
cy=\x2230.363636\x22 s\
odipodi:rx=\x2226.1\
81818\x22 sodipodi:\
ry=\x2226.181818\x22 d\
=\x22m 57.818182,30\
.363636 a 26.181\
818,26.181818 0 \
1 1 -52.363636,0\
 26.181818,26.18\
1818 0 1 1 52.36\
3636,0 z\x22 transf\
orm=\x22matrix(1.10\
76389,0,0,1.1076\
389,-3.0416672,9\
86.73024)\x22/>\x0a  <\
/g>\x0a</svg>\x0a\
\x00\x00\x16\xbd\
<\
?xml version=\x221.\
0\x22 encoding=\x22UTF\
-8\x22 standalone=\x22\
no\x22?>\x0a<!-- Creat\
ed with Inkscape\
 (http://www.ink\
scape.org/) -->\x0a\
\x0a<svg\x0a   xmlns:d\
c=\x22http://purl.o\
rg/dc/elements/1\
.1/\x22\x0a   xmlns:cc\
=\x22http://creativ\
ecommons.org/ns#\
\x22\x0a   xmlns:rdf=\x22\
http://www.w3.or\
g/1999/02/22-rdf\
-syntax-ns#\x22\x0a   \
xmlns:svg=\x22http:\
//www.w3.org/200\
0/svg\x22\x0a   xmlns=\
\x22http://www.w3.o\
rg/2000/svg\x22\x0a   \
xmlns:xlink=\x22htt\
p://www.w3.org/1\
999/xlink\x22\x0a   xm\
lns:sodipodi=\x22ht\
tp://sodipodi.so\
urceforge.net/DT\
D/sodipodi-0.dtd\
\x22\x0a   xmlns:inksc\
ape=\x22http://www.\
inkscape.org/nam\
espaces/inkscape\
\x22\x0a   width=\x2264\x22\x0a\
   height=\x2264\x22\x0a \
  id=\x22svg2\x22\x0a   v\
ersion=\x221.1\x22\x0a   \
inkscape:version\
=\x220.48.5 r10040\x22\
\x0a   sodipodi:doc\
name=\x22preview-ve\
ctor.svg\x22>\x0a  <de\
fs\x0a     id=\x22defs\
4\x22>\x0a    <inkscap\
e:perspective\x0a  \
     sodipodi:ty\
pe=\x22inkscape:per\
sp3d\x22\x0a       ink\
scape:vp_x=\x22-206\
.25075 : 119.688\
54 : 1\x22\x0a       i\
nkscape:vp_y=\x220 \
: 1000 : 0\x22\x0a    \
   inkscape:vp_z\
=\x22183.50001 : 14\
0.32143 : 1\x22\x0a   \
    inkscape:per\
sp3d-origin=\x2248 \
: 32 : 1\x22\x0a      \
 id=\x22perspective\
3792\x22 />\x0a    <li\
nearGradient\x0a   \
    id=\x22linearGr\
adient3755\x22>\x0a   \
   <stop\x0a       \
  style=\x22stop-co\
lor:#424242;stop\
-opacity:1;\x22\x0a   \
      offset=\x220\x22\
\x0a         id=\x22st\
op3757\x22 />\x0a     \
 <stop\x0a         \
style=\x22stop-colo\
r:#9f9f9f;stop-o\
pacity:1;\x22\x0a     \
    offset=\x221\x22\x0a \
        id=\x22stop\
3759\x22 />\x0a    </l\
inearGradient>\x0a \
   <linearGradie\
nt\x0a       inksca\
pe:collect=\x22alwa\
ys\x22\x0a       xlink\
:href=\x22#linearGr\
adient3755\x22\x0a    \
   id=\x22linearGra\
dient3761\x22\x0a     \
  x1=\x2275.724594\x22\
\x0a       y1=\x2285.0\
23956\x22\x0a       x2\
=\x2241.895164\x22\x0a   \
    y2=\x2213.33485\
1\x22\x0a       gradie\
ntUnits=\x22userSpa\
ceOnUse\x22 />\x0a  </\
defs>\x0a  <sodipod\
i:namedview\x0a    \
 id=\x22base\x22\x0a     \
pagecolor=\x22#ffff\
ff\x22\x0a     borderc\
olor=\x22#666666\x22\x0a \
    borderopacit\
y=\x221.0\x22\x0a     ink\
scape:pageopacit\
y=\x220.0\x22\x0a     ink\
scape:pageshadow\
=\x222\x22\x0a     inksca\
pe:zoom=\x228.75711\
68\x22\x0a     inkscap\
e:cx=\x2239.859065\x22\
\x0a     inkscape:c\
y=\x2230.925664\x22\x0a  \
   inkscape:docu\
ment-units=\x22px\x22\x0a\
     inkscape:cu\
rrent-layer=\x22g37\
94\x22\x0a     showgri\
d=\x22true\x22\x0a     sh\
owborder=\x22true\x22\x0a\
     inkscape:wi\
ndow-width=\x22800\x22\
\x0a     inkscape:w\
indow-height=\x2283\
6\x22\x0a     inkscape\
:window-x=\x220\x22\x0a  \
   inkscape:wind\
ow-y=\x2227\x22\x0a     i\
nkscape:window-m\
aximized=\x220\x22\x0a   \
  inkscape:snap-\
bbox=\x22false\x22\x0a   \
  inkscape:snap-\
nodes=\x22true\x22\x0a   \
  inkscape:snap-\
global=\x22true\x22>\x0a \
   <inkscape:gri\
d\x0a       type=\x22x\
ygrid\x22\x0a       id\
=\x22grid3002\x22\x0a    \
   empspacing=\x222\
\x22\x0a       visible\
=\x22true\x22\x0a       e\
nabled=\x22true\x22\x0a  \
     snapvisible\
gridlinesonly=\x22t\
rue\x22 />\x0a  </sodi\
podi:namedview>\x0a\
  <metadata\x0a    \
 id=\x22metadata7\x22>\
\x0a    <rdf:RDF>\x0a \
     <cc:Work\x0a  \
       rdf:about\
=\x22\x22>\x0a        <dc\
:format>image/sv\
g+xml</dc:format\
>\x0a        <dc:ty\
pe\x0a           rd\
f:resource=\x22http\
://purl.org/dc/d\
cmitype/StillIma\
ge\x22 />\x0a        <\
dc:title />\x0a    \
    <dc:creator>\
\x0a          <cc:A\
gent>\x0a          \
  <dc:title>$com\
mitter</dc:title\
>\x0a          </cc\
:Agent>\x0a        \
</dc:creator>\x0a  \
      <dc:title>\
preview-vector</\
dc:title>\x0a      \
  <dc:date>2013-\
11-19</dc:date>\x0a\
        <dc:rela\
tion>http://www.\
freecad.org/wiki\
/index.php?title\
=Artwork</dc:rel\
ation>\x0a        <\
dc:publisher>\x0a  \
        <cc:Agen\
t>\x0a            <\
dc:title>FreeCAD\
</dc:title>\x0a    \
      </cc:Agent\
>\x0a        </dc:p\
ublisher>\x0a      \
  <dc:identifier\
>FreeCAD/src/Mod\
/Material/Resour\
ces/icons/previe\
w-vector.svg</dc\
:identifier>\x0a   \
     <dc:rights>\
\x0a          <cc:A\
gent>\x0a          \
  <dc:title>Free\
CAD LGPL2+</dc:t\
itle>\x0a          \
</cc:Agent>\x0a    \
    </dc:rights>\
\x0a        <cc:lic\
ense>https://www\
.gnu.org/copylef\
t/lesser.html</c\
c:license>\x0a     \
   <dc:contribut\
or>\x0a          <c\
c:Agent>\x0a       \
     <dc:title>[\
agryson] Alexand\
er Gryson</dc:ti\
tle>\x0a          <\
/cc:Agent>\x0a     \
   </dc:contribu\
tor>\x0a      </cc:\
Work>\x0a    </rdf:\
RDF>\x0a  </metadat\
a>\x0a  <g\x0a     ink\
scape:label=\x22Lay\
er 1\x22\x0a     inksc\
ape:groupmode=\x22l\
ayer\x22\x0a     id=\x22l\
ayer1\x22\x0a     tran\
sform=\x22translate\
(0,-988.36218)\x22>\
\x0a    <g\x0a       i\
d=\x22g3794\x22\x0a      \
 style=\x22color:#0\
00000;fill:#ffff\
ff;fill-opacity:\
1;fill-rule:even\
odd;stroke:none;\
stroke-width:3.2\
9399991;marker:n\
one;visibility:v\
isible;display:i\
nline;overflow:v\
isible;enable-ba\
ckground:accumul\
ate\x22>\x0a      <pat\
h\x0a         style\
=\x22fill:#d3d7cf;f\
ill-opacity:1;fi\
ll-rule:evenodd;\
stroke:#000000;s\
troke-width:2;st\
roke-linecap:rou\
nd;stroke-linejo\
in:round;stroke-\
miterlimit:4;str\
oke-opacity:1;st\
roke-dasharray:n\
one\x22\x0a         d=\
\x22m 29.814326,991\
.36218 -8.814326\
,4 -1,7.00002 26\
,1 14.994091,-8.\
6577 z\x22\x0a        \
 id=\x22path3816\x22\x0a \
        inkscape\
:connector-curva\
ture=\x220\x22\x0a       \
  sodipodi:nodet\
ypes=\x22cccccc\x22 />\
\x0a      <path\x0a   \
      d=\x22M 38.48\
4871,1007.8171 4\
0,1046.3622 l 21\
,-22.4044 0,-29.\
25647 z\x22\x0a       \
  style=\x22fill:#b\
7b7b7;fill-opaci\
ty:1;fill-rule:e\
venodd;stroke:#0\
00000;stroke-wid\
th:2;stroke-line\
cap:round;stroke\
-linejoin:round;\
stroke-miterlimi\
t:4;stroke-opaci\
ty:1;stroke-dash\
array:none\x22\x0a    \
     id=\x22path381\
8\x22\x0a         inks\
cape:connector-c\
urvature=\x220\x22\x0a   \
      sodipodi:n\
odetypes=\x22ccccc\x22\
 />\x0a      <path\x0a\
         style=\x22\
fill:#555753;str\
oke:#555753;stro\
ke-width:2;strok\
e-linecap:butt;s\
troke-linejoin:m\
iter;stroke-opac\
ity:1\x22\x0a         \
d=\x22M 59.035745,9\
.6627807 59.1080\
06,34.818961 40.\
91531,54.264958 \
42.698526,19.372\
964 z\x22\x0a         \
id=\x22path3022\x22\x0a  \
       inkscape:\
connector-curvat\
ure=\x220\x22\x0a        \
 transform=\x22tran\
slate(0,988.3621\
8)\x22\x0a         sod\
ipodi:nodetypes=\
\x22ccccc\x22 />\x0a     \
 <path\x0a         \
d=\x22m 21,995.3621\
8 31,5.00002 -12\
,46 -30,-8 z\x22\x0a  \
       style=\x22fi\
ll:#ffffff;fill-\
opacity:1;fill-r\
ule:evenodd;stro\
ke:#000000;strok\
e-width:4;stroke\
-linejoin:round;\
stroke-miterlimi\
t:4;stroke-opaci\
ty:1;stroke-dash\
array:none\x22\x0a    \
     id=\x22path382\
0\x22\x0a         inks\
cape:connector-c\
urvature=\x220\x22\x0a   \
      sodipodi:n\
odetypes=\x22ccccc\x22\
 />\x0a      <path\x0a\
         d=\x22m 21\
,995.36218 31,5.\
00002 -12,46 -30\
,-8 z\x22\x0a         \
style=\x22color:#00\
0000;fill:none;s\
troke:#888a85;st\
roke-width:2;str\
oke-linejoin:rou\
nd;stroke-miterl\
imit:4;stroke-op\
acity:1;stroke-d\
asharray:none;ma\
rker:none;visibi\
lity:visible;dis\
play:inline;over\
flow:visible;ena\
ble-background:a\
ccumulate\x22\x0a     \
    id=\x22path3820\
-3\x22\x0a         ink\
scape:connector-\
curvature=\x220\x22\x0a  \
       sodipodi:\
nodetypes=\x22ccccc\
\x22 />\x0a    </g>\x0a  \
</g>\x0a</svg>\x0a\
"

qt_resource_name = b"\
\x00\x05\
\x00o\xa6S\
\x00i\
\x00c\x00o\x00n\x00s\
\x00\x14\
\x02\x9c\x05\x07\
\x00p\
\x00r\x00e\x00v\x00i\x00e\x00w\x00-\x00r\x00e\x00n\x00d\x00e\x00r\x00e\x00d\x00.\
\x00s\x00v\x00g\
\x00\x12\
\x0eO\xfeG\
\x00p\
\x00r\x00e\x00v\x00i\x00e\x00w\x00-\x00v\x00e\x00c\x00t\x00o\x00r\x00.\x00s\x00v\
\x00g\
"

qt_resource_struct = b"\
\x00\x00\x00\x00\x00\x02\x00\x00\x00\x01\x00\x00\x00\x01\
\x00\x00\x01\x8d\xe3s\x22p\
\x00\x00\x00\x00\x00\x02\x00\x00\x00\x02\x00\x00\x00\x02\
\x00\x00\x01\x8d\xe3s\x22p\
\x00\x00\x00\x10\x00\x00\x00\x00\x00\x01\x00\x00\x00\x00\
\x00\x00\x01\x8d\xe3s\x22p\
\x00\x00\x00>\x00\x00\x00\x00\x00\x01\x00\x00\x18\xa0\
\x00\x00\x01\x8d\xe3s\x22p\
"

def qInitResources():
    QtCore.qRegisterResourceData(0x03, qt_resource_struct, qt_resource_name, qt_resource_data)

def qCleanupResources():
    QtCore.qUnregisterResourceData(0x03, qt_resource_struct, qt_resource_name, qt_resource_data)

qInitResources()
